from datetime import date

from django.db import models
from textdistance import levenshtein


def levenshtein_ratio(s, t):
    return 1 - (levenshtein(s, t) / ((len(s) + len(t)) / 2))


class ShortHandManager(models.Manager):
    def get_similar(self, label):
        candidates = [(c, levenshtein_ratio(label, c.label)) for c in self.all()]
        candidates.sort(key=lambda x: x[1], reverse=True)
        candidates = [candidate[0].id for candidate in candidates if candidate[1] > .5]
        return self.filter(id__in=candidates)

    def get(self, *args, **kwargs):
        obj = super().get(*args, **kwargs)
        count_instance, _ = DailyCountInstance.objects.get_or_create(
            shorthand=obj,
            day=date.today()
        )
        count_instance.counter += 1
        count_instance.save(update_fields=['counter'])
        return obj


# Create your models here.
class ShortHand(models.Model):
    label = models.CharField(max_length=32, unique=True)
    url = models.URLField()
    objects = ShortHandManager()

    class Meta:
        ordering = ('-id',)

    def __str__(self):
        return f'{self.label} - {self.url}'

    @property
    def stats(self):
        return {c.day: c.counter for c in DailyCountInstance.objects.filter(shorthand=self)}


class DailyCountInstance(models.Model):
    shorthand = models.ForeignKey(ShortHand, on_delete=models.CASCADE)
    day = models.DateField()
    counter = models.IntegerField(default=0)

    class Meta:
        unique_together = ('shorthand', 'day',)

    def __str__(self):
        return f'{self.shorthand}: {self.day} :{self.counter}'



from rest_framework import serializers

from .models import ShortHand


class ShortHandSerializer(serializers.ModelSerializer):
    class Meta:
        model = ShortHand
        fields = ['id', 'label', 'url',]


from django.urls import path

from . import views

urlpatterns = [
    path('', views.index, name='index'),
    path('<str:shorthand>', views.reroute, name='index'),
    path('<str:shorthand>/<str:parameter>', views.reroute, name='index')
]


from django.test import TestCase

from backend.models import levenshtein_ratio


class TestLevenshteinRatio(TestCase):
    def test_distance_same_str(self):
        self.assertEqual(1, levenshtein_ratio('foo', 'foo'))

    def test_distance_similar_str(self):
        self.assertEqual(.6, levenshtein_ratio('foo', 'fo'))

    def test_distance_very_different_str(self):
        self.assertEqual(0, levenshtein_ratio('foo', 'bar'))



from django.urls import include, path
from rest_framework import routers

from . import views

router = routers.DefaultRouter()
router.register(r'shorthand', views.ShortHandViewSet)

urlpatterns = [
    path('', include(router.urls))
]



from django.shortcuts import render, redirect
from django.http.response import HttpResponseRedirect

from backend.models import ShortHand


def index(request):
    if request.method == 'POST' and (request.POST.get('shorthand') and request.POST.get('url')):
        ShortHand.objects.create(
            label=request.POST.get('shorthand'),
            url=request.POST.get('url')
        )
        return redirect('index')

    return render(request, 'index.html', context={'shorthands': ShortHand.objects.all()})


def reroute(request, shorthand, parameter=None):
    qs = ShortHand.objects.filter(label=shorthand)
    if not qs:
        qs = ShortHand.objects.get_similar(shorthand)
        return render(request, '404.html', status=404, context={'shorthand': shorthand, 'candidates': qs})

    obj = ShortHand.objects.get(label=shorthand)
    url = obj.url
    if not parameter and '{}' in url:
        url = url.split('{}')[0]
    elif parameter:
        if '{}' in url:
            url = url.format(parameter)
        else:
            url = f'{url.rstrip("/")}/{parameter}'
    return HttpResponseRedirect(url)



# Generated by Django 3.0.6 on 2020-06-02 03:40

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='ShortHand',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('label', models.CharField(max_length=32, unique=True)),
                ('url', models.URLField()),
            ],
        ),
    ]



# Generated by Django 3.0.6 on 2020-06-09 02:14

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('backend', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='DailyCountInstance',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('day', models.DateField()),
                ('counter', models.IntegerField(default=0)),
                ('shorthand', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='backend.ShortHand')),
            ],
            options={
                'unique_together': {('shorthand', 'day')},
            },
        ),
    ]



from django.test import Client
from django.test import TestCase

from backend.models import ShortHand


class TestCreate(TestCase):
    def setUp(self):
        self.client = Client()

    def test_creating_shorthand_from_url(self):
        self.assertFalse(ShortHand.objects.filter(url='http://bar.com', label='foo').exists())
        response = self.client.post('/', {'shorthand': 'foo', 'url': 'http://bar.com'})
        self.assertTrue(ShortHand.objects.filter(url='http://bar.com', label='foo').exists())




from rest_framework import viewsets

from .models import ShortHand
from .serializers import ShortHandSerializer


class ShortHandViewSet(viewsets.ModelViewSet):
    queryset = ShortHand.objects.all()
    serializer_class = ShortHandSerializer



import os
from setuptools import setup, find_packages

path = os.path.dirname(__file__)
long_desc_fd = os.path.join(path, 'README.md')

long_description = ''

with open(long_desc_fd) as f:
    long_description = f.read()

setup(
    name='gollama',
    version='1.1.0',
    packages=find_packages(),
    url='https://github.com/sillyfatcat/gollama',
    license='MIT',
    author='Shelby Shum',
    author_email='sshum00@gmail.com',
    description='Django implementation for golink',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=['golink', 'tinyurl', 'shorthand'],
    download_url='https://github.com/sillyfatcat/gollama/archive/v0.1.tar.gz',
    install_requires=[
        'django',
        'djangorestframework',
        'textdistance',
    ],
    include_package_data=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Developers',
        'Topic :: Internet :: Name Service (DNS)',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)



from django.test import TestCase
from rest_framework.test import APIClient


class TestShortHand(TestCase):
    def setUp(self) -> None:
        self.client = APIClient()

    def test_get(self):
        response_json = self.client.post('/api/v1/shorthand/', {'label': 'foo', 'url': 'http://bar.com'}).json()
        response = self.client.get(f'/api/v1/shorthand/{response_json["id"]}/')
        self.assertEqual(200, response.status_code)
        self.assertEqual({
            'id': 1,
            'label': 'foo',
            'url': 'http://bar.com'
        }, response.json())

    def test_list(self):
        self.client.post('/api/v1/shorthand/', {'label': 'foo', 'url': 'http://bar.com'})
        self.client.post('/api/v1/shorthand/', {'label': 'bar', 'url': 'http://foo.com'})
        response = self.client.get('/api/v1/shorthand/')
        self.assertEqual(200, response.status_code)
        response_json = response.json()
        self.assertEqual(2, len(response_json))

    def test_create(self):
        response = self.client.post('/api/v1/shorthand/', {'label': 'foo', 'url': 'http://bar.com'})
        self.assertEqual(201, response.status_code)
        response_json = response.json()
        self.assertEqual({
            'id': 1,
            'label': 'foo',
            'url': 'http://bar.com'
        }, response_json)

    def test_create_fail_duplicate(self):
        response = self.client.post('/api/v1/shorthand/', {'label': 'foo', 'url': 'http://bar.com'})
        self.assertEqual(201, response.status_code)
        response = self.client.post('/api/v1/shorthand/', {'label': 'foo', 'url': 'http://bar.com'})
        self.assertEqual(400, response.status_code)

    def test_update(self):
        response = self.client.post('/api/v1/shorthand/', {'label': 'foo', 'url': 'http://bar.com'})
        self.assertEqual(201, response.status_code)
        response_json = response.json()
        self.assertEqual('http://bar.com', response_json['url'])
        response = self.client.patch(f'/api/v1/shorthand/{response_json["id"]}/', {'url': 'https://bar.com'})
        self.assertEqual(200, response.status_code)
        response_json = response.json()
        self.assertEqual('https://bar.com', response_json['url'])

    def test_delete(self):
        response = self.client.post('/api/v1/shorthand/', {'label': 'foo', 'url': 'http://bar.com'})
        self.assertEqual(201, response.status_code)
        response_json = response.json()
        self.assertEqual({
            'id': 1,
            'label': 'foo',
            'url': 'http://bar.com'
        }, response_json)
        response = self.client.delete(f'/api/v1/shorthand/{response_json["id"]}/')
        self.assertEqual(204, response.status_code)
        response = self.client.get(f'/api/v1/shorthand/{response_json["id"]}/')
        self.assertEqual(404, response.status_code)



from django.db.utils import IntegrityError
from django.test import TestCase

from backend.serializers import ShortHandSerializer


class TestShortHand(TestCase):
    def test_validate_true(self):
        serialized = ShortHandSerializer(data={'label': 'foo', 'url': 'http://foo.com'})
        self.assertTrue(serialized.is_valid())

    def test_validate_false(self):
        serialized = ShortHandSerializer(data={'url': 'http://foo.com'})
        self.assertFalse(serialized.is_valid())
        serialized = ShortHandSerializer(data={'label': 'foo'})
        self.assertFalse(serialized.is_valid())

    def test_create(self):
        serialized = ShortHandSerializer(data={'label': 'foo', 'url': 'http://foo.com'})
        serialized.is_valid()
        obj = serialized.save()
        self.assertIsNotNone(obj)
        self.assertEqual(1, obj.id)
        self.assertEqual('foo', obj.label)
        self.assertEqual('http://foo.com', obj.url)

    def test_create_fail_duplicate(self):
        serialized = ShortHandSerializer(data={'label': 'foo', 'url': 'http://foo.com'})
        serialized.is_valid()
        obj = serialized.save()
        self.assertIsNotNone(obj)
        serialized = ShortHandSerializer(data={'label': 'foo', 'url': 'http://foo.com'})
        self.assertFalse(serialized.is_valid())

    def test_update(self):
        serialized = ShortHandSerializer(data={'label': 'foo', 'url': 'http://foo.com'})
        serialized.is_valid()
        obj = serialized.save()
        serialized = ShortHandSerializer(obj, {'url': 'http://bar.com'}, partial=True)
        self.assertTrue(serialized.is_valid())
        obj = serialized.save()
        self.assertEqual('http://bar.com', obj.url)

    def test_json(self):
        serialized = ShortHandSerializer(data={'label': 'foo', 'url': 'http://foo.com'})
        serialized.is_valid()
        obj = serialized.save()
        self.assertEqual({
            'id': 1,
            'label': 'foo',
            'url': 'http://foo.com',
        }, ShortHandSerializer(obj).data)
